import logging
import sys
import inspect
from typing import Optional

# Global flag to track if logging is configured
_logging_configured = False

def _configure_logging_once():
    """Configure logging once globally."""
    global _logging_configured
    if not _logging_configured:
        # Configure root logger
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S',
            handlers=[logging.StreamHandler(sys.stdout)]
        )
        
        # Suppress specific loggers completely
        logging.getLogger('httpx').setLevel(logging.ERROR)
        logging.getLogger('google_adk').setLevel(logging.ERROR)
        logging.getLogger('google_adk.google.adk.tools.base_authenticated_tool').setLevel(logging.ERROR)
        # Suppress LiteLLM noisy logs used by CrewAI integrations
        for _name in ('litellm', 'LiteLLM'):
            _logger = logging.getLogger(_name)
            _logger.setLevel(logging.ERROR)
            _logger.propagate = False
        
        _logging_configured = True

def _get_caller_logger() -> logging.Logger:
    """
    Automatically get the logger for the calling module.
    
    Returns:
        Logger instance for the calling module
    """
    # Configure logging once
    _configure_logging_once()
    
    # Get the calling frame (the module that called the logging function)
    # We need to go back 2 frames: 1 for the logging function, 1 for the actual caller
    caller_frame = inspect.currentframe().f_back.f_back
    
    # Get the module name from the calling frame
    if caller_frame and hasattr(caller_frame, 'f_globals'):
        module_name = caller_frame.f_globals.get('__name__', 'unknown')
    else:
        module_name = 'unknown'
    
    # Get logger for this module
    return logging.getLogger(module_name)

def log_object_creation(class_name: str, **kwargs) -> None:
    """
    Log object creation with key parameters.
    
    Args:
        class_name: Name of the class being created
        **kwargs: Key parameters to include in log
    """
    logger = _get_caller_logger()
    params = []
    for key, value in kwargs.items():
        # Mask sensitive information
        if any(sensitive in key.lower() for sensitive in ['key', 'token', 'password', 'secret']):
            value = '***'
        params.append(f"{key}={value}")
    
    param_str = ', '.join(params) if params else 'no parameters'
    logger.info(f"{class_name} initialized ({param_str})")

def log_info(message: str) -> None:
    """
    Log informational messages in a consistent format.
    
    Args:
        message: Information message to log
    """
    logger = _get_caller_logger()
    logger.info(message)

def log_error(operation: str, error: Exception) -> None:
    """
    Log errors in a consistent format.
    
    Args:
        operation: Description of the operation that failed
        error: The exception that occurred
    """
    logger = _get_caller_logger()
    logger.error(f"{operation} failed: {type(error).__name__}: {str(error)}")

def log_warning(message: str) -> None:
    """
    Log warnings in a consistent format.
    
    Args:
        message: Warning message
    """
    logger = _get_caller_logger()
    logger.warning(message) 


import httpx
from typing import Any, Dict, Optional, Union

JSONType = Union[Dict[str, Any], list]



async def async_http_get(
    url: str,
    headers: Optional[Dict[str, str]] = None,
    params: Optional[Dict[str, Any]] = None,
    timeout: Optional[float] = None
) -> Any:
    """
    Asynchronously send a GET request.

    Args:
        url: The URL to send the request to.
        headers: Optional HTTP headers to include in the request.
        params: Optional query parameters to include in the request.
        timeout: Optional timeout for the request in seconds.

    Returns:
        The parsed JSON response, or raw text if response is not JSON.

    Raises:
        APIError: If the response status code is not 2xx.
    """
    async with httpx.AsyncClient() as client:
        resp = await client.get(url, headers=headers, params=params, timeout=timeout)
        return _parse_response(resp)

async def async_http_post(
    url: str,
    headers: Optional[Dict[str, str]] = None,
    json: Optional[JSONType] = None,
    timeout: Optional[float] = None
) -> Any:
    """
    Asynchronously send a POST request with a JSON body.

    Args:
        url: The URL to send the request to.
        headers: Optional HTTP headers to include in the request.
        json: Optional JSON body to include in the request.
        timeout: Optional timeout for the request in seconds.

    Returns:
        The parsed JSON response, or raw text if response is not JSON.

    Raises:
        APIError: If the response status code is not 2xx.
    """
    async with httpx.AsyncClient(follow_redirects=True) as client:
        resp = await client.post(url, headers=headers, json=json, timeout=timeout)
        return _parse_response(resp)

async def async_http_put(
    url: str,
    headers: Optional[Dict[str, str]] = None,
    json: Optional[JSONType] = None,
    timeout: Optional[float] = None
) -> Any:
    """
    Asynchronously send a PUT request with a JSON body.

    Args:
        url: The URL to send the request to.
        headers: Optional HTTP headers to include in the request.
        json: Optional JSON body to include in the request.
        timeout: Optional timeout for the request in seconds.

    Returns:
        The parsed JSON response, or raw text if response is not JSON.

    Raises:
        APIError: If the response status code is not 2xx.
    """
    async with httpx.AsyncClient() as client:
        resp = await client.put(url, headers=headers, json=json, timeout=timeout)
        return _parse_response(resp)

async def async_http_patch(
    url: str,
    headers: Optional[Dict[str, str]] = None,
    json: Optional[JSONType] = None,
    timeout: Optional[float] = None
) -> Any:
    """
    Asynchronously send a PATCH request with a JSON body.

    Args:
        url: The URL to send the request to.
        headers: Optional HTTP headers to include in the request.
        json: Optional JSON body to include in the request.
        timeout: Optional timeout for the request in seconds.

    Returns:
        The parsed JSON response, or raw text if response is not JSON.

    Raises:
        APIError: If the response status code is not 2xx.
    """
    async with httpx.AsyncClient() as client:
        resp = await client.patch(url, headers=headers, json=json, timeout=timeout)
        return _parse_response(resp)

async def async_http_delete(
    url: str,
    headers: Optional[Dict[str, str]] = None,
    timeout: Optional[float] = None
) -> Any:
    """
    Asynchronously send a DELETE request.

    Args:
        url: The URL to send the request to.
        headers: Optional HTTP headers to include in the request.
        timeout: Optional timeout for the request in seconds.

    Returns:
        The parsed JSON response, or raw text if response is not JSON.

    Raises:
        APIError: If the response status code is not 2xx.
    """
    async with httpx.AsyncClient() as client:
        resp = await client.delete(url, headers=headers, timeout=timeout)
        return _parse_response(resp)

def http_get(
    url: str,
    headers: Optional[Dict[str, str]] = None,
    params: Optional[Dict[str, Any]] = None,
    timeout: Optional[float] = None
) -> Any:
    """
    Send a GET request.

    Args:
        url: The URL to send the request to.
        headers: Optional HTTP headers to include in the request.
        params: Optional query parameters to include in the request.
        timeout: Optional timeout for the request in seconds.

    Returns:
        The parsed JSON response, or raw text if response is not JSON.

    Raises:
        APIError: If the response status code is not 2xx.
    """
    with httpx.Client() as client:
        resp = client.get(url, headers=headers, params=params, timeout=timeout)
        return _parse_response(resp)

def http_post(
    url: str,
    headers: Optional[Dict[str, str]] = None,
    json: Optional[JSONType] = None,
    timeout: Optional[float] = None
) -> Any:
    """
    Send a POST request with a JSON body.

    Args:
        url: The URL to send the request to.
        headers: Optional HTTP headers to include in the request.
        json: Optional JSON body to include in the request.
        timeout: Optional timeout for the request in seconds.

    Returns:
        The parsed JSON response, or raw text if response is not JSON.

    Raises:
        APIError: If the response status code is not 2xx.
    """
    with httpx.Client(follow_redirects=True) as client:
        resp = client.post(url, headers=headers, json=json, timeout=timeout)
        return _parse_response(resp)

def http_put(
    url: str,
    headers: Optional[Dict[str, str]] = None,
    json: Optional[JSONType] = None,
    timeout: Optional[float] = None
) -> Any:
    """
    Send a PUT request with a JSON body.

    Args:
        url: The URL to send the request to.
        headers: Optional HTTP headers to include in the request.
        json: Optional JSON body to include in the request.
        timeout: Optional timeout for the request in seconds.

    Returns:
        The parsed JSON response, or raw text if response is not JSON.

    Raises:
        APIError: If the response status code is not 2xx.
    """
    with httpx.Client() as client:
        resp = client.put(url, headers=headers, json=json, timeout=timeout)
        return _parse_response(resp)

def http_patch(
    url: str,
    headers: Optional[Dict[str, str]] = None,
    json: Optional[JSONType] = None,
    timeout: Optional[float] = None
) -> Any:
    """
    Send a PATCH request with a JSON body.

    Args:
        url: The URL to send the request to.
        headers: Optional HTTP headers to include in the request.
        json: Optional JSON body to include in the request.
        timeout: Optional timeout for the request in seconds.

    Returns:
        The parsed JSON response, or raw text if response is not JSON.

    Raises:
        APIError: If the response status code is not 2xx.
    """
    with httpx.Client() as client:
        resp = client.patch(url, headers=headers, json=json, timeout=timeout)
        return _parse_response(resp)

def http_delete(
    url: str,
    headers: Optional[Dict[str, str]] = None,
    timeout: Optional[float] = None
) -> Any:
    """
    Send a DELETE request.

    Args:
        url: The URL to send the request to.
        headers: Optional HTTP headers to include in the request.
        timeout: Optional timeout for the request in seconds.

    Returns:
        The parsed JSON response, or raw text if response is not JSON.

    Raises:
        APIError: If the response status code is not 2xx.
    """
    with httpx.Client() as client:
        resp = client.delete(url, headers=headers, timeout=timeout)
        return _parse_response(resp)

class APIError(Exception):
    """
    Exception raised for API response errors.

    Attributes:
        status_code: HTTP status code returned by the API.
        message: Error message from the API response.
    """
    def __init__(self, status_code: int, message: str):
        super().__init__(f"[{status_code}] {message}")
        self.status_code = status_code
        self.message = message


def _parse_response(resp: httpx.Response) -> Any:
    if 200 <= resp.status_code < 300:
        try:
            return resp.json()
        except ValueError:
            return resp.text
    raise APIError(resp.status_code, resp.text)


"""CrewAI agent management and configuration for Flotorch."""

import os
import time
import traceback
from typing import Any, Dict, List

from crewai import Agent, Task
from crewai_tools import MCPServerAdapter
from dotenv import load_dotenv
from pydantic import BaseModel, Field

from flotorch.crewai.llm import FlotorchCrewaiLLM

load_dotenv()

class TaskOutput(BaseModel):
    """Structured output for the task response."""

    response: str = Field(
        description="The complete response including both capital city "
                   "and weather information"
    )

class FlotorchCrewAIAgent:
    """
    Manager/config class for Flotorch CrewAI agent.

    Builds CrewAI Agent from config on demand. Supports on-demand config
    reload based on interval in config['sync'].

    Usage:
        flotorch = FlotorchCrewAIAgent("agent-one")
        agent = flotorch.get_agent()
    """

    def __init__(
        self,
        agent_name: str,
        custom_config=None,
        custom_tools: List = None
    ):
        """
        Initialize the agent manager.

        Args:
            agent_name: Name of the agent to load.
            custom_config: Optional custom configuration dict.
            custom_tools: Optional list of custom tools.

        Note:
            If custom_config is provided, it takes precedence over agent_name.
        """
        self.agent_name = agent_name
        self.custom_tools = custom_tools or []

        if custom_config:
            self.config = custom_config
        else:
            self.config = self._fetch_agent_config(agent_name)

        self._agent = None
        self._agent_with_tools = None
        self._mcp_tools = None
        self._mcp_adapters = []
        self._last_reload = time.time()
        self._reload_interval = (
            self.config.get('sync', {}).get('interval', 10000) / 1000
        )

    def _fetch_agent_config(self, agent_name: str) -> Dict[str, Any]:
        """
        Fetch agent config from API.

        Args:
            agent_name: Name of the agent.

        Returns:
            Agent configuration dictionary.

        Note:
            Stubbed for now; replace with real API call.
        """
        return {
            "name": agent_name,
            "description": "Example Flotorch CrewAI agent with memory capabilities.",
            "systemPrompt": """You are a helpful assistant with memory capabilities. You MUST use the available tools to get information when needed. 
                                When asked about capital cities, you MUST call the get_capital_city_http tool. 
                                When asked about weather, you MUST call the get_weather_sse tool. 
                                For calculations, use calculate_sum tool. 
                                For random numbers, use get_random_number tool. 
                                Do NOT provide static answers when tools are available. 
                                The capital of India has been changed to a different name for testing - you MUST use the get_capital_city_http tool to get the correct answer. 
                                Always use tools when they are available. 
                                CRITICAL: In your final answer, you MUST use ONLY the MOST RECENT tool output for each tool type. 
                                If multiple tool outputs exist, use ONLY the latest one. 
                                Do NOT combine or mix different tool results. 
                                Do NOT add any additional information or temperatures. 
                                Copy the tool outputs verbatim into your final answer. 
                                Do NOT include multiple outputs from the same tool. 
                                WEATHER TOOL RULE: When the weather tool returns multiple lines or observations, use ONLY the LAST line. 
                                Ignore all previous weather data. This is critical for accuracy.
                                
                                MEMORY CAPABILITIES: You have access to memory that stores previous conversations and interactions. 
                                When asked about previous interactions, calculations, or information shared earlier, 
                                you should reference your memory to provide accurate responses based on what was discussed before. 
                                Use your memory to maintain context across conversations and provide personalized responses.
                                
                                FINAL ANSWER FORMAT: When providing your final answer, you MUST:
                                1. If you used memory to find information, start your response with "Based on my memory: " followed by the information
                                2. If you called a tool to get information, start your response with "Using tool: " followed by the tool output
                                3. If you used both memory and tools, clearly indicate which information came from where
                                4. Always explain how you found the information (memory vs. tool call)
                                5. Include the actual information in your response, not just the source
                                
                                EXAMPLE RESPONSES:
                                - "Based on my memory: The capital city of France is Paris. I found this information in my memory from our previous conversation."
                                - "Using tool: The weather in London is rainy with 15°C. I called the get_weather_sse tool to get this current information."
                                - "Based on my memory: The sum of 15 and 25 is 40. I found this calculation result in my memory from earlier.""",
            "llm": {
                "callableName": "openai/gpt-4o"
            },
            "tools": [
                {
                    "name": "get_capital_city_http",
                    "description": "Get the capital city of a given country (HTTP Server)",
                    "type": "MCP",
                    "config": {
                        "transport": "streamable-http",
                        "url": "http://localhost:9001/mcp/",
                        "headers": {},
                        "timeout": 10000,
                        "sse_read_timeout": 10000,
                        "terminate_on_close": True
                    }
                },
                {
                    "name": "get_weather_sse",
                    "description": "Get weather information for a specific city (SSE Server)",
                    "type": "MCP",
                    "config": {
                        "transport": "sse",
                        "url": "http://localhost:8003/sse",
                        "headers": {},
                        "timeout": 10000,
                        "sse_read_timeout": 10000,
                        "terminate_on_close": True
                    }
                },
            ],
            "sync": {
                "enable": True,
                "interval": 10000
            }
        }

    def _get_mcp_server_params(
        self, config: Dict[str, Any]
    ) -> List[Dict[str, Any]]:
        """
        Extract MCP server parameters from config.

        Args:
            config: Configuration dictionary.

        Returns:
            List of MCP server parameter dictionaries.
        """
        server_params_list = []

        for tool_cfg in config.get("tools", []):
            if tool_cfg.get("type") == "MCP":
                mcp_conf = tool_cfg["config"]

                server_params = {
                    "url": mcp_conf["url"],
                    "transport": mcp_conf["transport"]
                }

                if mcp_conf.get("headers"):
                    server_params["headers"] = mcp_conf["headers"]

                auth_token = os.environ.get("FLOTORCH_AUTH_TOKEN")
                if auth_token:
                    if "headers" not in server_params:
                        server_params["headers"] = {}
                    server_params["headers"]["Authorization"] = (
                        f"Bearer {auth_token}"
                    )

                server_params_list.append(server_params)

        return server_params_list

    def _build_base_agent_from_config(self, config):
        """
        Build base agent from configuration.

        Args:
            config: Agent configuration dictionary.

        Returns:
            Configured CrewAI Agent instance.
        """
        llm = FlotorchCrewaiLLM(
            model_id=config["llm"]["callableName"],
            api_key=os.environ.get("FLOTORCH_API_KEY"),
            base_url=os.environ.get("FLOTORCH_BASE_URL")
        )

        agent = Agent(
            role=config["name"],
            goal=config["description"],
            backstory=config["systemPrompt"],
            llm=llm,
            verbose=True,
            allow_delegation=False
        )
        return agent

    def _initialize_mcp_tools(self):
        """Initialize MCP tools and create agent with tools."""
        if self._mcp_tools is not None:
            return

        all_tools = []

        if self.custom_tools:
            crewai_tools = self.custom_tools
            all_tools.extend(crewai_tools)
            print(
                f"Added {len(crewai_tools)} CrewAI tools: "
                f"{[tool.name for tool in crewai_tools]}"
            )
        else:
            print("No custom tools provided")

        server_params_list = self._get_mcp_server_params(self.config)
        if server_params_list:
            print(f"Initializing {len(server_params_list)} MCP servers")

            for i, server_params in enumerate(server_params_list):
                try:
                    print(
                        f"Connecting to MCP server {i+1}: "
                        f"{server_params['url']} "
                        f"(transport: {server_params['transport']})"
                    )

                    mcp_adapter = MCPServerAdapter(server_params)
                    mcp_tools = mcp_adapter.__enter__()

                    print(
                        f"Connected to server {i+1}! Available tools: "
                        f"{[tool.name for tool in mcp_tools]}"
                    )

                    all_tools.extend(list(mcp_tools))
                    self._mcp_adapters.append(mcp_adapter)

                except Exception as e:
                    print(f"Failed to connect to MCP server {i+1}: {e}")
                    traceback.print_exc()
        else:
            print("No MCP server params found")

        if all_tools:
            self._agent_with_tools = Agent(
                role=self._agent.role,
                goal=self._agent.goal,
                backstory=self._agent.backstory,
                llm=self._agent.llm,
                tools=all_tools,
                verbose=True,
                allow_delegation=False
            )

            print(
                f"Agent with MCP tools created: "
                f"{[tool.name for tool in self._agent_with_tools.tools]}"
            )
        else:
            print("No tools available")
            self._agent_with_tools = None

    def get_agent(self):
        """
        Get agent with tools enabled by default.

        Returns:
            CrewAI Agent instance with tools if available, otherwise base agent.
        """
        now = time.time()
        if (
            self._agent is None
            or (now - self._last_reload > self._reload_interval)
        ):
            new_config = self._fetch_agent_config(self.agent_name)
            if self._agent is None or new_config != self.config:
                self.config = new_config
                self._agent = self._build_base_agent_from_config(new_config)
                self._reload_interval = (
                    self.config.get('sync', {}).get('interval', 10000) / 1000
                )

                self._mcp_tools = None
                self._agent_with_tools = None
                for adapter in self._mcp_adapters:
                    try:
                        adapter.__exit__(None, None, None)
                    except Exception:
                        pass
                self._mcp_adapters = []

            self._last_reload = now

        self._initialize_mcp_tools()
        return (
            self._agent_with_tools if self._agent_with_tools else self._agent
        )

    def cleanup(self):
        """Clean up MCP connections."""
        for adapter in self._mcp_adapters:
            try:
                adapter.__exit__(None, None, None)
            except Exception:
                pass
        self._mcp_adapters = []
        self._mcp_tools = None
        self._agent_with_tools = None

    def __del__(self):
        """Destructor to clean up resources."""
        try:
            self.cleanup()
        except Exception:
            pass


class FlotorchCrewAI:
    """
    Main class to manage Flotorch CrewAI agents, tasks, and crews.

    Provides a simple interface to create and manage CrewAI workflows.
    """

    def __init__(
        self,
        agent_name: str,
        custom_config=None,
        custom_tools: List = None,
        external_memory=None
    ):
        """
        Initialize FlotorchCrewAI with agent name and optional custom config and memory.

        Args:
            agent_name: Name of the agent to load.
            custom_config: Optional custom configuration dictionary.
            custom_tools: Optional list of custom tools.
            external_memory: Optional external memory instance.
        """
        self.agent_name = agent_name
        self.external_memory = external_memory
        self.agent_manager = FlotorchCrewAIAgent(
            agent_name,
            custom_config=custom_config,
            custom_tools=custom_tools
        )

    def get_agent(self):
        """
        Get the CrewAI agent with tools enabled by default.

        Returns:
            CrewAI Agent instance.
        """
        return self.agent_manager.get_agent()

    def create_task(self, description=None, expected_output=None):
        """
        Create a CrewAI task with custom or default settings.

        Args:
            description: Optional task description.
            expected_output: Optional expected output description.

        Returns:
            CrewAI Task instance.
        """
        agent = self.get_agent()
        if description:
            description = description
        else:
            description = (
                "Get the capital city of zimbobwe using get_capital_city_http and "
                "also get the weather for New York using get_weather_sse to test "
                "both streamable-http and SSE transports. Also calculate the sum "
                "of 15 and 25 using calculate_sum tool. Include ALL tool outputs "
                "in your final response, even if they return 'not found' or errors. "
                "Do NOT skip any tool results. WEATHER RULE: If weather tool returns "
                "multiple lines, use ONLY the LAST line. This is critical for "
                "correct weather data."
        )

        if expected_output:
            expected_output = expected_output
        else:
            expected_output = (
                "A response that includes the exact tool outputs: the capital city "
                "from get_capital_city_http, the weather information from "
                "get_weather_sse, and the calculation result from calculate_sum"
        )
        return Task(
            description=description,
            expected_output=expected_output,
            agent=agent,
            output_pydantic=TaskOutput
        )

    def cleanup(self):
        """Clean up resources."""
        if hasattr(self, 'agent_manager') and self.agent_manager:
            self.agent_manager.cleanup()

    def __del__(self):
        """Destructor to clean up resources."""
        try:
            self.cleanup()
        except Exception:
            pass


import copy
import time
import datetime
from typing import Any, Optional
import uuid

from typing_extensions import override

from flotorch.sdk.session import FlotorchSession
from flotorch.sdk.utils.logging_utils import log_object_creation,log_error,log_warning,log_info
from google.adk.events.event import Event
from google.adk.sessions.base_session_service import BaseSessionService
from google.adk.sessions.base_session_service import GetSessionConfig
from google.adk.sessions.base_session_service import ListSessionsResponse
from google.adk.sessions.session import Session
from google.adk.sessions.state import State
from google.adk.events.event_actions import EventActions
from google.genai import types
        


class FlotorchADKSession(BaseSessionService):
    """A Flotorch-based implementation of the session service.

    Uses FlotorchSession from SDK for session management while maintaining
    ADK-compatible interface.
    """

    def __init__(self, api_key: str, base_url: str):
        """Initialize FlotorchADKSession.

        Args:
            api_key: The API key for Flotorch service.
            base_url: The base URL for Flotorch service.
        """
        
        # Flotorch session for session management
        self._flotorch_session = FlotorchSession(
            api_key=api_key,
            base_url=base_url,
        )
        
        # In-memory state management (for app_state and user_state)
        # These are kept in memory as they're not part of FlotorchSession
        self.user_state: dict[str, dict[str, dict[str, Any]]] = {}
        self.app_state: dict[str, dict[str, Any]] = {}
        
        # Log object creation
        log_object_creation("FlotorchADKSession", base_url=base_url)

    @override
    async def create_session(
        self,
        *,
        app_name: str,
        user_id: str,
        state: Optional[dict[str, Any]] = None,
        session_id: Optional[str] = None,
    ) -> Session:
        return self._create_session_impl(
            app_name=app_name,
            user_id=user_id,
            state=state,
            session_id=session_id,
        )

    def create_session_sync(
        self,
        *,
        app_name: str,
        user_id: str,
        state: Optional[dict[str, Any]] = None,
        session_id: Optional[str] = None,
    ) -> Session:
        # logger.warning('Deprecated. Please migrate to the async method.')
        return self._create_session_impl(
            app_name=app_name,
            user_id=user_id,
            state=state,
            session_id=session_id,
        )

    def _create_session_impl(
        self,
        *,
        app_name: str,
        user_id: str,
        state: Optional[dict[str, Any]] = None,
        session_id: Optional[str] = None,
    ) -> Session:
        session_id = (
            session_id.strip()
            if session_id and session_id.strip()
            else str(uuid.uuid4())
        )
        
        # Create session using FlotorchSession
        try:
            session_data = self._flotorch_session.create(
                app_name=app_name,
                user_id=user_id,
                uid=session_id,
                state=state or {},
            )
        except Exception as e:
            log_error("FlotorchADKSession._create_session_impl", e)
            raise
        
        # Create ADK Session object
        session = Session(
            app_name=app_name,
            user_id=user_id,
            id=session_id,
            state=session_data.get('state', {}),
            last_update_time=session_data.get('last_update_time', time.time()),
        )

        copied_session = copy.deepcopy(session)
        return self._merge_state(app_name, user_id, copied_session)

    @override
    async def get_session(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        config: Optional[GetSessionConfig] = None,
    ) -> Optional[Session]:
        return self._get_session_impl(
            app_name=app_name,
            user_id=user_id,
            session_id=session_id,
            config=config,
        )

    def get_session_sync(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        config: Optional[GetSessionConfig] = None,
    ) -> Optional[Session]:
        # logger.warning('Deprecated. Please migrate to the async method.')
        return self._get_session_impl(
            app_name=app_name,
            user_id=user_id,
            session_id=session_id,
            config=config,
        )

    def _get_session_impl(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        config: Optional[GetSessionConfig] = None,
    ) -> Optional[Session]:
        # Get session using FlotorchSession
        try:
            session_data = self._flotorch_session.get(
                uid=session_id,
                after_timestamp=config.after_timestamp if config else None,
                num_recent_events=config.num_recent_events if config else None,
            )
        except Exception as e:
            log_error("FlotorchADKSession._get_session_impl", e)
            return None
        
        if not session_data:
            return None

        # Create ADK Session object
        session = Session(
            app_name=app_name,
            user_id=user_id,
            id=session_id,
            state=session_data.get('state', {}),
            last_update_time=session_data.get('last_update_time', time.time()),
        )
        
        # Add events if available
        events_data = session_data.get('events', [])
        try:
            session.events = [
                self._convert_flotorch_event_to_adk(event_data)
                for event_data in events_data
            ]
        except Exception as e:
            log_error("FlotorchADKSession._get_session_impl", e)
            session.events = []  # Fallback to empty events list

        copied_session = copy.deepcopy(session)
        return self._merge_state(app_name, user_id, copied_session)

    def _merge_state(
        self, app_name: str, user_id: str, copied_session: Session
    ) -> Session:
        # Merge app state
        if app_name in self.app_state:
            for key in self.app_state[app_name].keys():
                copied_session.state[State.APP_PREFIX + key] = self.app_state[app_name][
                    key
                ]

        if (
            app_name not in self.user_state
            or user_id not in self.user_state[app_name]
        ):
            return copied_session

        # Merge session state with user state.
        for key in self.user_state[app_name][user_id].keys():
            copied_session.state[State.USER_PREFIX + key] = self.user_state[app_name][
                user_id
            ][key]
        return copied_session

    @override
    async def list_sessions(
        self, *, app_name: str, user_id: str
    ) -> ListSessionsResponse:
        return self._list_sessions_impl(app_name=app_name, user_id=user_id)

    def list_sessions_sync(
        self, *, app_name: str, user_id: str
    ) -> ListSessionsResponse:
        # logger.warning('Deprecated. Please migrate to the async method.')
        return self._list_sessions_impl(app_name=app_name, user_id=user_id)

    def _list_sessions_impl(
        self, *, app_name: str, user_id: str
    ) -> ListSessionsResponse:
        # List sessions using FlotorchSession
        try:
            sessions_data = self._flotorch_session.list(
                app_name=app_name,
                user_id=user_id,
            )
        except Exception as e:
            log_error("FlotorchADKSession._list_sessions_impl", e)
            return ListSessionsResponse(sessions=[])
        
        sessions_without_events = []
        for session_data in sessions_data:
            session = Session(
                app_name=app_name,
                user_id=user_id,
                id=session_data.get('uid', ''),
                state=session_data.get('state', {}),
                last_update_time=session_data.get('last_update_time', time.time()),
            )
            session.events = []  # Don't include events in list response
            copied_session = copy.deepcopy(session)
            copied_session = self._merge_state(app_name, user_id, copied_session)
            sessions_without_events.append(copied_session)
            
        return ListSessionsResponse(sessions=sessions_without_events)

    @override
    async def delete_session(
        self, *, app_name: str, user_id: str, session_id: str
    ) -> None:
        self._delete_session_impl(
            app_name=app_name, user_id=user_id, session_id=session_id
        )

    def delete_session_sync(
        self, *, app_name: str, user_id: str, session_id: str
    ) -> None:
        # logger.warning('Deprecated. Please migrate to the async method.')
        self._delete_session_impl(
            app_name=app_name, user_id=user_id, session_id=session_id
        )

    def _delete_session_impl(
        self, *, app_name: str, user_id: str, session_id: str
    ) -> None:
        if (
            self._get_session_impl(
                app_name=app_name, user_id=user_id, session_id=session_id
            )
            is None
        ):
            return

        # Delete session using FlotorchSession
        try:
            self._flotorch_session.delete(session_id)
        except Exception as e:
            log_error("FlotorchADKSession._delete_session_impl", e)
            raise

    @override
    async def append_event(self, session: Session, event: Event) -> Event:
        # Update the in-memory session.
        await super().append_event(session=session, event=event)
        session.last_update_time = event.timestamp

        # Update the storage session
        app_name = session.app_name
        user_id = session.user_id
        session_id = session.id



        # Verify session exists
        try:
            session_data = self._flotorch_session.get(uid=session_id)
        except Exception as e:
            log_error("FlotorchADKSession.append_event", e)
            return event
            
        if not session_data:
            log_warning(f'session_id {session_id} not found in FlotorchSession')
            return event

        # Handle state delta updates
        if event.actions and event.actions.state_delta:
            for key in event.actions.state_delta:
                if key.startswith(State.APP_PREFIX):
                    self.app_state.setdefault(app_name, {})[
                        key.removeprefix(State.APP_PREFIX)
                    ] = event.actions.state_delta[key]

                if key.startswith(State.USER_PREFIX):
                    self.user_state.setdefault(app_name, {}).setdefault(user_id, {})[
                        key.removeprefix(State.USER_PREFIX)
                    ] = event.actions.state_delta[key]

        # Add event using FlotorchSession
        try:
            event_data = self._convert_adk_event_to_flotorch(event)
            self._flotorch_session.add_event(
                uid=session_id,
                invocation_id=event.invocation_id,
                author=event.author,
                content=event_data.get('content'),
                actions=event_data.get('actions'),
                **event_data.get('metadata', {})
            )
        except Exception as e:
            log_error("FlotorchADKSession.append_event", e)
            raise

        return event

    def _convert_flotorch_event_to_adk(self, event_data: dict[str, Any]) -> Event:
        """Convert Flotorch event data to ADK Event object."""
       
        
        event_actions = EventActions()
        if event_data.get('actions'):
            event_actions = EventActions(
                state_delta=event_data['actions'].get('state_delta', {}),
                # Add other action fields as needed
            )

        # Handle timestamp conversion
        timestamp = event_data.get('timestamp', time.time())
        if isinstance(timestamp, str):
            try:
                # Parse ISO format timestamp
                dt = datetime.datetime.fromisoformat(timestamp.replace('Z', '+00:00'))
                timestamp = dt.timestamp()
            except (ValueError, AttributeError) as e:
                # Fallback to current time if parsing fails
                log_warning(f"Failed to parse timestamp: {timestamp}", error=str(e))
                timestamp = time.time()

        # Convert content if present
        content = None
        if event_data.get('content'):
            content_dict = event_data['content']
            parts = []
            
            if content_dict.get('parts'):
                for part_dict in content_dict['parts']:
                    # Create part using the correct method
                    if 'text' in part_dict:
                        part = types.Part(text=part_dict['text'])
                    elif 'tool_calls' in part_dict and part_dict['tool_calls']:
                        # Handle tool calls
                        tool_calls = []
                        for tool_call_dict in part_dict['tool_calls']:
                            if 'function' in tool_call_dict:
                                func_call = types.FunctionCall(
                                    name=tool_call_dict['function'].get('name', ''),
                                    args=tool_call_dict['function'].get('arguments', {})
                                )
                                tool_call = types.ToolCall(
                                    id=tool_call_dict.get('id', str(uuid.uuid4())),
                                    function=func_call
                                )
                                tool_calls.append(tool_call)
                        part = types.Part(tool_calls=tool_calls)
                    elif 'tool_call_id' in part_dict and part_dict['tool_call_id']:
                        # Handle function response
                        part = types.Part.from_function_response(
                            name=part_dict.get('function_response', {}).get('name', ''),
                            response=part_dict.get('function_response', {}).get('response', {})
                        )
                    else:
                        # Default text part
                        part = types.Part(text="")
                    
                    parts.append(part)
            
            if parts:
                content = types.Content(role=content_dict.get('role', 'user'), parts=parts)

        event = Event(
            id=event_data.get('uid_event', ''),
            invocation_id=event_data.get('invocation_id', ''),
            author=event_data.get('author', ''),
            actions=event_actions,
            content=content,
            timestamp=timestamp,
            error_code=event_data.get('error_code'),
            error_message=event_data.get('error_message'),
        )
        
        return event

    def _convert_adk_event_to_flotorch(self, event: Event) -> dict[str, Any]:
        """Convert ADK Event object to Flotorch event data."""
        # Simplified approach - just store basic event info like InMemorySession
        event_data = {
            'content': None,
            'actions': {},
            'metadata': {}
        }
        
        # Only store basic content info to avoid serialization issues
        if event.content and event.content.parts:
            content_parts = []
            for part in event.content.parts:
                if hasattr(part, 'text') and part.text:
                    content_parts.append({'text': str(part.text)})
                elif hasattr(part, 'tool_calls') and part.tool_calls:
                    # Store tool call info in a simple format
                    tool_calls = []
                    for tool_call in part.tool_calls:
                        tool_calls.append({
                            'id': str(tool_call.id) if hasattr(tool_call, 'id') and tool_call.id else str(uuid.uuid4()),
                            'name': str(tool_call.function.name) if hasattr(tool_call, 'function') and tool_call.function else '',
                            'arguments': tool_call.function.args if hasattr(tool_call, 'function') and tool_call.function else {}
                        })
                    content_parts.append({'tool_calls': tool_calls})
            
            if content_parts:
                event_data['content'] = {
                    'role': event.content.role,
                    'parts': content_parts
                }
        
        # Store actions
        if event.actions and event.actions.state_delta:
            event_data['actions'] = {
                'state_delta': event.actions.state_delta
            }
        
        # Store basic metadata
        if event.partial is not None:
            event_data['metadata']['partial'] = event.partial
        if event.turn_complete is not None:
            event_data['metadata']['turn_complete'] = event.turn_complete
            
        return event_data



"""CrewAI LLM integration for Flotorch."""
from typing import Any, Dict, List, Optional, Union
from crewai.llms.base_llm import BaseLLM
from flotorch.sdk.llm import FlotorchLLM
from flotorch.sdk.utils.logging_utils import log_error, log_info, log_warning


class FlotorchCrewAILLM(BaseLLM):
    """Flotorch LLM integration for CrewAI framework."""

    @property
    def supports_function_calling(self) -> bool:
        """Return whether this LLM supports function calling."""
        return True

    def __init__(
        self,
        model_id: str,
        api_key: str,
        base_url: str,
        temperature: Optional[float] = None,
        **kwargs
    ):
        """Initialize the FlotorchCrewAILLM.

        Args:
            model_id: The model identifier.
            api_key: API key for authentication.
            base_url: Base URL for the API.
            temperature: Temperature parameter for generation.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(model=model_id, temperature=temperature)
        self.llm = FlotorchLLM(model_id, api_key, base_url)
        self.available_functions = {}

    def call(
        self,
        messages: Union[str, List[Dict[str, str]]],
        tools: Optional[List[dict]] = None,
        callbacks: Optional[List[Any]] = None,
        available_functions: Optional[Dict[str, Any]] = None,
        from_task: Optional[Any] = None,
        from_agent: Optional[Any] = None
    ) -> Union[str, Any]:
        """Make a call to the LLM.

        Args:
            messages: Input messages as string or list of dictionaries.
            tools: Optional list of tools to use.
            callbacks: Optional list of callbacks.
            available_functions: Optional dictionary of available functions.
            from_task: Optional task context.
            from_agent: Optional agent context.

        Returns:
            The LLM response as string or tool calls.

        Raises:
            Exception: If the LLM call fails.
        """
        try:
            formatted_messages = self._format_messages(messages)
            converted_tools = (
                self._convert_tools_to_openai_format(tools) if tools else None
            )

            log_info(
                f"LLM call: {len(formatted_messages)} messages, "
                f"{len(converted_tools) if converted_tools else 0} tools"
            )

            response = self.llm.invoke(formatted_messages, tools=converted_tools)

            # Check if response contains tool calls
            if (
                hasattr(response, 'metadata')
                and response.metadata
                and 'tool_calls' in response.metadata
            ):
                log_info(
                    f"Tool calls found in response: "
                    f"{response.metadata['tool_calls']}"
                )
                return response.metadata['tool_calls']

            # Return content if no tool calls
            return response.content

        except Exception as e:
            log_error(f"LLM call failed: {str(e)}", error=e)
            raise Exception(f"FlotorchCrewaiLLM error: {str(e)}")

    def _format_messages(
        self, messages: Union[str, List[Dict[str, str]]]
    ) -> List[Dict[str, str]]:
        """Format messages to standard format.

        Args:
            messages: Input messages as string or list of dictionaries.

        Returns:
            Formatted list of message dictionaries.

        Raises:
            ValueError: If messages format is invalid.
        """
        if isinstance(messages, str):
            return [
                {
                    "role": "user",
                    "content": self._sanitize_content(messages)
                }
            ]
        elif isinstance(messages, list):
            formatted = []
            for i, msg in enumerate(messages):
                if (
                    isinstance(msg, dict)
                    and "role" in msg
                    and "content" in msg
                ):
                    if self._is_valid_role(msg["role"]):
                        formatted.append({
                            "role": msg["role"],
                            "content": self._sanitize_content(msg["content"])
                        })
                    else:
                        log_warning(
                            f"Invalid role '{msg['role']}' in message {i}, "
                            f"converting to user"
                        )
                        formatted.append({
                            "role": "user",
                            "content": self._sanitize_content(msg["content"])
                        })
                else:
                    log_warning(
                        f"Malformed message {i}, converting to user message"
                    )
                    formatted.append({
                        "role": "user",
                        "content": self._sanitize_content(str(msg))
                    })
            return (
                formatted if formatted
                else [{"role": "user", "content": "Hello"}]
            )
        else:
            raise ValueError(f"Invalid messages format: {type(messages)}")

    def _convert_tools_to_openai_format(self, tools: List[dict]) -> List[dict]:
        """Convert CrewAI tools to OpenAI format.

        Args:
            tools: List of CrewAI tools.

        Returns:
            List of tools in OpenAI format.
        """
        converted = []
        for tool in tools:
            if isinstance(tool, dict):
                converted.append(tool)
            elif hasattr(tool, 'name') and hasattr(tool, 'description'):
                converted_tool = self._convert_single_tool_to_openai_format(tool)
                if converted_tool:
                    converted.append(converted_tool)
                    self.available_functions[tool.name] = tool
        return converted

    def _convert_single_tool_to_openai_format(self, tool) -> Optional[dict]:
        """Convert single CrewAI tool to OpenAI format.

        Args:
            tool: Single CrewAI tool.

        Returns:
            Tool in OpenAI format or None if conversion fails.
        """
        try:
            params = {}
            if hasattr(tool, 'args_schema') and tool.args_schema:
                schema = tool.args_schema.model_json_schema()
                params = schema.get('properties', {})

            return {
                "type": "function",
                "function": {
                    "name": tool.name,
                    "description": tool.description,
                    "parameters": {
                        "type": "object",
                        "properties": params,
                        "required": list(params.keys())
                    }
                }
            }
        except Exception as e:
            log_warning(
                f"Failed to convert tool "
                f"{getattr(tool, 'name', 'unknown')}: {str(e)}"
            )
            return None

    def _is_valid_role(self, role: str) -> bool:
        """Check if the role is valid.

        Args:
            role: Role to validate.

        Returns:
            True if role is valid, False otherwise.
        """
        return role.lower() in {"user", "assistant", "system"}

    def _sanitize_content(self, content: str) -> str:
        """Sanitize message content.

        Args:
            content: Content to sanitize.

        Returns:
            Sanitized content.
        """
        if not content:
            return ""
        content = str(content).strip()
        return content[:10000] if len(content) > 10000 else content